# Generated by Django 3.1.2 on 2020-11-02 17:48

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0019_auto_20201030_1913'),
    ]

    operations = [
        migrations.AddField(
            model_name='election',
            name='remind_text_sent',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.1.2 on 2020-11-16 20:28

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0023_election_disable_abstention'),
    ]

    operations = [
        migrations.AlterField(
            model_name='election',
            name='result_published',
            field=models.CharField(choices=[('0', 'unpublished'), ('1', 'published')], default='0', max_length=1),
        ),
    ]



import csv
import io
import re
from datetime import timedelta
from typing import Tuple, List

from django import forms
from django.conf import settings
from django.core.validators import validate_email
from django.utils import timezone

from management.models import ElectionManager
from vote.models import Election, Application, Session, Voter, OpenVote


class StartElectionForm(forms.ModelForm):
    run_time = forms.IntegerField(label="run time", min_value=1)

    class Meta:
        model = Election
        fields: List[str] = []

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.start_date = timezone.now()
        instance.end_date = timezone.now() + timedelta(minutes=self.cleaned_data['run_time'])
        if commit:
            instance.save()

        return instance


class StopElectionForm(forms.ModelForm):
    class Meta:
        model = Election
        fields: List[str] = []

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.end_date = timezone.now()
        if commit:
            instance.save()

        return instance


class TemplateStringForm:
    def clean_email_text(self, test_data: List[str], field: str):
        """
        checks if the cleaned_text fails when formatted on the test_data.
        """
        test_data_dict = {i: "" for i in test_data}
        cleaned_text = self.cleaned_data[field]
        try:
            cleaned_text.format(**test_data_dict)
        except (KeyError, ValueError, IndexError):
            x = re.findall(r"\{\w*\}", cleaned_text)
            test_data = set(x) - {f"{{{i}}}" for i in test_data}
            self.add_error(field, "The following variables are not allowed: " + ", ".join(test_data))
        return cleaned_text


class AddSessionForm(forms.ModelForm, TemplateStringForm):
    variables = {
        "{name}": "Voter's name if set",
        "{title}": "Session's title",
        "{access_code}": "Access code/token for the voter to login",
        "{login_url}": "URL which instantly logs user in",
        "{base_url}": f"Will render to: https://{settings.URL}",
        "{start_time}": "Start time if datetime is set",
        "{start_date}": "Start date if datetime is set",
        "{start_time_en}": "Start time in english format e.g. 02:23 PM",
        "{start_date_en}": "Start date in english format e.g. 12/12/2020",
        "{meeting_link}": "Meeting link if set"
    }

    email = forms.EmailField(required=False, label="",
                             widget=forms.EmailInput(attrs={"class": "emailinput form-control",
                                                            "placeholder": "your@email.de"}))

    def __init__(self, request, user, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['start_date'].widget = forms.TextInput(attrs={'placeholder': 'e.g. 2020-05-12 13:00:00',
                                                                  'type': 'datetime'})
        self.fields['meeting_link'].widget = forms.TextInput(
            attrs={'placeholder': 'e.g. https://bbb.stusta.de/b/ssn-abc-123'})
        self.user = user
        self.request = request

    class Meta:
        model = Session
        fields = ('title', 'start_date', 'meeting_link', 'invite_text')
        labels = {
            'title': 'Session Name',
            'start_date': 'Meeting start (optional)',
            'meeting_link': 'Link to meeting call platform (optional)',
            # will be set by html
            'invite_text': ''
        }

    def clean(self):
        super().clean()
        if self.request.POST.get("submit_type") == "test" and not self.cleaned_data['email']:
            self.add_error('email', "Email must be set for sending the test mail.")
        if self.request.POST.get("submit_type") == "test" and not self.cleaned_data['invite_text']:
            self.add_error('invite_text', "The test email can only be send when the invite text is filled.")
        return self.cleaned_data

    def clean_invite_text(self):
        test_data = ["name", "title", "access_code", "login_url", "base_url", "start_time",
                     "start_date", "meeting_link", "start_date_en", "start_time_en"]
        return self.clean_email_text(test_data, 'invite_text')

    def _save_m2m(self):
        super()._save_m2m()

        if not self.user.sessions.filter(pk=self.instance.pk):
            self.user.sessions.add(self.instance)
            self.user.save()

    def save(self, commit=True):
        self.instance = super().save(commit=False)

        if commit:
            self.instance.save()
            self._save_m2m()
        else:
            self.save_m2m = self._save_m2m  # pylint: disable=attribute-defined-outside-init

        return self.instance


class SessionSettingsForm(AddSessionForm):
    add_election_manager = forms.CharField(max_length=256, required=False, label='')

    class Meta:
        model = Session
        fields = ('start_date', 'meeting_link', 'invite_text')
        labels = {
            'start_date': 'Meeting start (optional)',
            'meeting_link': 'Link to meeting call platform (optional)',
            # will be set by html
            'invite_text': ''
        }
        widgets = {
            'start_date': forms.TextInput(attrs={'placeholder': 'e.g. 2020-05-12 13:00:00', 'type': 'datetime'})
        }

    def clean_add_election_manager(self):
        value = self.data['add_election_manager']
        if not value:
            return value
        if not ElectionManager.objects.filter(username=value).exists():
            raise forms.ValidationError(f'Cannot find election manager with username {value}')

        return ElectionManager.objects.get(username=value)

    def _save_m2m(self):
        super()._save_m2m()

        if self.cleaned_data['add_election_manager']:
            self.cleaned_data['add_election_manager'].sessions.add(self.instance)
            self.cleaned_data['add_election_manager'].save()

    def save(self, commit=True):
        self.instance = super().save(commit=False)
        if commit:
            self.instance.save()
            self._save_m2m()
        else:
            self.save_m2m = self._save_m2m  # pylint: disable=attribute-defined-outside-init

        return self.instance


class AddElectionForm(forms.ModelForm, TemplateStringForm):
    variables = {
        "{name}": "Voter's name if set",
        "{title}": "Session's title",
        "{url}": "URL to the election",
        "{end_time}": "End time if datetime is set",
        "{end_date}": "End date if datetime is set",
        "{end_time_en}": "End time in english format e.g. 02:23 PM",
        "{end_date_en}": "End date in english format e.g. 12/12/2020",
    }

    email = forms.EmailField(required=False, label="",
                             widget=forms.EmailInput(attrs={"class": "emailinput form-control",
                                                            "placeholder": "your@email.de"}))

    def __init__(self, user, session, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['session'].disabled = True
        self.fields['session'].initial = session
        self.fields['session'].widget = forms.HiddenInput()
        self.fields['start_date'].widget = forms.TextInput(
            attrs={'placeholder': 'e.g.: 2020-05-12 13:00:00', 'type': 'datetime'})
        self.fields['end_date'].widget = forms.TextInput(
            attrs={'placeholder': 'e.g.: 2020-05-12 13:00:00', 'type': 'datetime'})
        # self.fields['start_date'].initial = timezone.now()
        self.fields['max_votes_yes'] = forms.IntegerField(min_value=1, required=False,
                                                          label='Maximum number of YES votes (optional)')
        self.user = user
        self.session = session
        self.request = request

    class Meta:
        model = Election
        fields = (
            'title', 'start_date', 'end_date', 'session', 'max_votes_yes', 'voters_self_apply', 'send_emails_on_start',
            'remind_text', 'enable_abstention', 'result_published')

        labels = {
            'title': 'Election Name',
            'start_date': 'Start time (optional)',
            'end_date': 'End time (optional)',
            'voters_self_apply': 'Voters can apply for the election',
            'send_emails_on_start': 'Voters receive an e-mail when the election starts<br>'
                                    '(useful for elections that last several days)',
            'enable_abstention': 'Enable the option to abstain in this election<br>'
                                 '(YES,  NO and ABSTENTION votes will be allowed)',
            'remind_text': '',
            'result_published': 'Automatically publish the election results',
        }

    def clean_remind_text(self):
        test_data = ["name", "title", "url", "end_time", "end_date", "end_date_en", "end_time_en"]
        return self.clean_email_text(test_data, 'remind_text')

    def clean(self):
        super().clean()
        if self.session not in self.user.sessions.all():
            raise forms.ValidationError("You don't have the permission to add an election here.")
        if not self.cleaned_data['send_emails_on_start'] and self.cleaned_data['remind_text']:
            self.add_error('send_emails_on_start', "Remind text can only be set when this option is activated.")
        if self.request.POST.get("submit_type") == "test" and not self.cleaned_data['email']:
            self.add_error('email', "Email must be set for sending the test mail.")
        if self.request.POST.get("submit_type") == "test" and not self.cleaned_data['remind_text']:
            self.add_error('remind_text', "The test email can only be send when the remind text is filled.")
        if self.cleaned_data['end_date'] and self.cleaned_data['end_date'] < timezone.now():
            self.add_error('end_date', "End date cannot be in the past.")
        if self.cleaned_data.get('end_date') and self.cleaned_data.get('start_date') and \
                self.cleaned_data['start_date'] > self.cleaned_data['end_date']:
            raise forms.ValidationError("Start date needs to be before end date")
        return self.cleaned_data

    def save(self, commit=True):
        instance = super().save(commit=commit)
        self.session.elections.add(instance)
        if commit:
            self.session.save()
            open_votes = [
                OpenVote(voter=v, election=instance)
                for v in self.session.participants.all()
            ]
            OpenVote.objects.bulk_create(open_votes)

        return instance


class AvatarFileInput(forms.ClearableFileInput):
    template_name = 'management/image_input.html'


class ApplicationUploadForm(forms.ModelForm):
    field_order = ['election', 'display_name', 'email', 'text', 'avatar']

    def __init__(self, election, request, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request = request
        self.election = election
        self.fields['election'].required = False

    class Meta:
        model = Application
        fields = ('election', 'display_name', 'email', 'text', 'avatar')

    def clean(self):
        super().clean()
        if not self.election.can_apply:
            raise forms.ValidationError('Applications are currently not allowed')
        return self.cleaned_data

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.election = self.election
        if commit:
            instance.save()

        return instance


class AddVotersForm(forms.Form):
    voters_list = forms.CharField(widget=forms.Textarea, label='Emails')  # explicitly no max_length here

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    def save(self) -> List[Tuple[Voter, str]]:
        voters_codes = [
            Voter.from_data(email=email, session=self.session) for email in self.cleaned_data['voters_list']
        ]
        self.session.managers.all().first().send_invite_bulk_threaded(voters_codes)
        return voters_codes

    def clean_voters_list(self):
        lines = self.cleaned_data['voters_list'].splitlines()
        emails = []
        for line in lines:
            if line == '':
                continue

            try:
                validate_email(line)
            except forms.ValidationError:
                self.add_error('voters_list', f'{line} is not a valid email address')

            if Voter.objects.filter(email=line, session=self.session).exists():
                self.add_error('voters_list', f'a voter with email address {line} already exists')

            emails.append(line)

        if len(emails) != len(set(emails)):
            raise forms.ValidationError('duplicate email address')

        return emails


class AddTokensForm(forms.Form):
    nr_anonymous_voters = forms.IntegerField(min_value=1, max_value=50, label='Number of Tokens:')

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    def save(self) -> List[Tuple[Voter, str]]:
        anonymous_voters = [
            Voter.from_data(session=self.session) for _ in range(self.cleaned_data['nr_anonymous_voters'])
        ]

        return anonymous_voters


class CSVUploaderForm(forms.Form):
    csv_data = forms.FileField(label='CSV File')

    def __init__(self, session, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.session = session

    def clean_csv_data(self):
        f = self.cleaned_data['csv_data'].file
        data = {}
        try:
            with io.TextIOWrapper(f, encoding='utf-8') as text_file:
                csv_reader = csv.DictReader(text_file)
                if 'email' not in csv_reader.fieldnames or 'name' not in csv_reader.fieldnames:
                    raise forms.ValidationError('CSV file needs to have columns "email" and "name".')
                for row in csv_reader:
                    if row['email']:
                        try:
                            validate_email(row['email'])
                        except forms.ValidationError:
                            self.add_error('csv_data', f'Invalid email {row["email"]}')
                    else:
                        row['email'] = None

                    if Voter.objects.filter(session=self.session, email=row['email']).exists():
                        self.add_error('csv_data', f'Voter with email address {row["email"]} already exists')

                    if row['email'] in data:
                        self.add_error('csv_data', f'Duplicate email in csv: {row["email"]}')

                    data[row['email']] = row['name']
        except UnicodeDecodeError as e:
            raise forms.ValidationError('File does not seem to be in CSV format.') from e

        return data

    def save(self):
        voters_codes = [
            Voter.from_data(session=self.session, email=email, name=name)
            for email, name in self.cleaned_data['csv_data'].items()
        ]
        self.session.managers.all().first().send_invite_bulk_threaded(voters_codes)



from django.conf import settings


def is_valid_sender_email(email: str) -> bool:
    if not isinstance(email, str) or '@' not in email:
        return False

    return email.split('@')[-1] in settings.VALID_MANAGER_EMAIL_DOMAINS



# Generated by Django 3.0.6 on 2020-05-29 12:52

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import vote.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Application',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.TextField(blank=True, max_length=250)),
                ('avatar', models.ImageField(blank=True, null=True, upload_to=vote.models.avatar_file_name)),
                ('last_name', models.CharField(max_length=256)),
                ('first_name', models.CharField(max_length=256)),
                ('room', models.CharField(blank=True, max_length=64)),
                ('email', models.EmailField(blank=True, max_length=254)),
            ],
        ),
        migrations.CreateModel(
            name='Election',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=512)),
                ('start_date', models.DateTimeField(blank=True, null=True)),
                ('end_date', models.DateTimeField(blank=True, null=True)),
                ('max_votes_yes', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=256)),
                ('meeting_link', models.CharField(blank=True, max_length=512, null=True)),
                ('start_date', models.DateTimeField(blank=True, default=django.utils.timezone.now, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='Voter',
            fields=[
                ('voter_id', models.IntegerField(primary_key=True, serialize=False)),
                ('password', models.CharField(max_length=256)),
                ('first_name', models.CharField(max_length=128)),
                ('last_name', models.CharField(max_length=128)),
                ('email', models.EmailField(max_length=254)),
                ('remind_me', models.BooleanField(default=False)),
                ('session', models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='participants', to='vote.Session')),
            ],
        ),
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('vote', models.CharField(choices=[('abstention', 'Abstention'), ('accept', 'Yes'), ('reject', 'No')], max_length=10)),
                ('candidate', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='votes', to='vote.Application')),
                ('election', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='votes', to='vote.Election')),
            ],
        ),
        migrations.CreateModel(
            name='OpenVote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('election', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='open_votes', to='vote.Election')),
                ('voter', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='vote.Voter')),
            ],
        ),
        migrations.AddField(
            model_name='election',
            name='session',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='elections', to='vote.Session'),
        ),
        migrations.AddField(
            model_name='application',
            name='election',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='application', to='vote.Election'),
        ),
    ]



from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.decorators import user_passes_test
from django_auth_ldap.backend import LDAPBackend

from management.models import ElectionManager


def management_login_required(function=None, redirect_field_name=REDIRECT_FIELD_NAME, login_url='management:login'):
    """
    Decorator for views that checks that the voter is logged in, redirecting
    to the log-in page if necessary.
    """
    actual_decorator = user_passes_test(
        lambda u: u.is_authenticated and isinstance(u, ElectionManager),
        login_url=login_url,
        redirect_field_name=redirect_field_name
    )
    if function:
        return actual_decorator(function)
    return actual_decorator


class ManagementBackend(BaseBackend):
    def authenticate(self, request, **kwargs):
        username = kwargs.pop('username', None)
        password = kwargs.pop('password', None)

        if username is None or password is None:
            return None

        try:
            user = ElectionManager.objects.get(username=username)
        except ElectionManager.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            ElectionManager().set_password(password)
        else:
            if user.check_password(password):
                user.backend = 'management.authentication.ManagementBackend'
                return user

        return None

    def get_user(self, user_id):
        return ElectionManager.objects.filter(pk=user_id).first()


class ManagementBackendLDAP(LDAPBackend):
    def get_user_model(self):
        return ElectionManager



# Generated by Django 3.1.4 on 2020-12-08 21:45

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0024_auto_20201116_2128'),
    ]

    operations = [
        migrations.AddField(
            model_name='session',
            name='spectator_token',
            field=models.UUIDField(blank=True, null=True, unique=True),
        ),
    ]



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserCreationForm, UserChangeForm
from django.utils.translation import gettext_lazy as _

from .models import Application
from .models import Election
from .models import Voter


class VoterCreationForm(UserCreationForm):
    class Meta:
        model = Voter
        fields = ('voter_id',)
        field_classes = {}


class VoterChangeForm(UserChangeForm):
    class Meta:
        model = Voter
        fields = ('voter_id',)
        field_classes = {}


class VoterAdmin(UserAdmin):
    add_form = VoterCreationForm
    form = VoterChangeForm
    model = Voter
    fieldsets = (
        (None, {'fields': ('voter_id', 'password', 'session',)}),
        (_('Personal info'), {'fields': ('email',)}),
        (_('Status'), {'fields': ('voted',)}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('voter_id', 'password1', 'password2', 'session')}
         ),
        (_('Personal info'), {'fields': ('email',)}),
    )
    list_display = ('voter_id', 'session',)
    list_filter = ()
    search_fields = ('voter_id', 'session', 'email',)
    ordering = ('voter_id',)
    filter_horizontal = tuple()


admin.site.register(Election)
admin.site.register(Application)
admin.site.register(Voter, VoterAdmin)



from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils import timezone

from vote.models import Session


class Command(BaseCommand):
    help = 'Creates a new election'

    def add_arguments(self, parser):
        parser.add_argument('-t', '--title', type=str, required=True)

        # Make things a little bit easier for dev and debugging convenience
        if settings.DEBUG:
            parser.add_argument('-s', '--start-date', type=str, default=timezone.now())
            parser.add_argument('-l', '--meeting-link', type=str, default="http://meeting.link")
            parser.add_argument('-id', type=int, default=0)
        else:
            parser.add_argument('-s', '--start-date', type=str, required=True)
            parser.add_argument('-l', '--meeting-link', type=str, required=True)

    def handle(self, *args, **options):
        session = Session.objects.create(
            id=options['id'],
            title=options['title'],
            start_date=options['start_date'],
            meeting_link=options['meeting_link'],
        )
        self.stdout.write(self.style.SUCCESS('Successfully created session "%s" with ID %i' % (session, session.id)))



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys
from pathlib import Path


def setup():
    """Setup environment for wahlfang"""
    if os.getenv('DJANGO_SETTINGS_MODULE') is not None:
        return

    if os.getenv('WAHLFANG_DEBUG'):
        os.environ['DJANGO_SETTINGS_MODULE'] = 'wahlfang.settings.development'
        return

    wahlfang_config = os.getenv('WAHLFANG_CONFIG', '/etc/wahlfang/settings.py')
    if not os.path.exists(wahlfang_config):
        print(f'Wahlfang configuration file at {wahlfang_config} does not exist', file=sys.stderr)
        print('Modify "WAHLFANG_CONFIG" environment variable to point at settings.py', file=sys.stderr)
        sys.exit(1)

    config_path = Path(wahlfang_config).resolve()
    sys.path.append(str(config_path.parent))

    os.environ['DJANGO_SETTINGS_MODULE'] = config_path.stem


def main():
    setup()

    try:
        from django.core.management import execute_from_command_line  # pylint: disable=import-outside-toplevel
    except ImportError as exc:
        raise ImportError(
            "Couldn't import Django. Are you sure it's installed and "
            "available on your PYTHONPATH environment variable? Did you "
            "forget to activate a virtual environment?"
        ) from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()



# Generated by Django 3.0.5 on 2020-05-29 19:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0002_auto_20200529_2104'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='application',
            name='room',
        ),
        migrations.AlterField(
            model_name='application',
            name='email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
    ]



from django.contrib.auth import views as auth_views
from django.urls import path
from django.views.generic.base import RedirectView

from vote import views

app_name = 'vote'

urlpatterns = [
    path('', views.index, name='index'),
    path('vote/<int:election_id>', views.vote, name='vote'),

    # code login
    path('code', views.LoginView.as_view(), name='code_login'),
    path('code/', RedirectView.as_view(pattern_name='code_login')),
    path('code/<str:access_code>', views.code_login, name='link_login'),
    path('logout', auth_views.LogoutView.as_view(
        next_page='vote:index',
    ), name='logout'),
    path('vote/<int:election_id>/apply', views.apply, name='apply'),
    path('vote/<int:election_id>/delete-own-application', views.delete_own_application, name='delete_own_application'),
    path('help', views.help_page, name='help'),
    path('spectator/<uuid:uuid>', views.spectator, name='spectator')
]



# Generated by Django 3.1.8 on 2021-05-01 16:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0025_session_spectator_token'),
    ]

    operations = [
        migrations.AddField(
            model_name='voter',
            name='invalid_email',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.1.13 on 2021-08-04 21:35

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0027_change_field_type_results_published'),
    ]

    operations = [
        migrations.AlterField(
            model_name='application',
            name='election',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='applications', to='vote.election'),
        ),
        migrations.AlterField(
            model_name='application',
            name='voter',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='applications', to='vote.voter'),
        ),
    ]



# Generated by Django 3.1.4 on 2021-06-18 13:47

from django.db import migrations, models


def char2bool(apps, schema_editor):
    election = apps.get_model('vote', 'election')
    for row in election.objects.all():
        row.result_unpublished = not bool(int(row.result_published))
        row.save(update_fields=['result_unpublished'])


def bool2char(apps, schema_editor):
    election = apps.get_model('vote', 'election')
    for row in election.objects.all():
        row.result_published = str(int(not row.result_unpublished))
        row.save(update_fields=['result_published'])


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0026_voter_invalid_email'),
    ]

    operations = [
        migrations.AddField(
            model_name='election',
            name='result_unpublished',
            field=models.BooleanField(default=True),
        ),
        migrations.RunPython(char2bool, reverse_code=bool2char),
        migrations.RemoveField(
            model_name='election',
            name='result_published'
        )
    ]



"""
ASGI config for wahlfang project.

It exposes the ASGI callable as a module-level variable named ``application``.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/howto/deployment/asgi/
"""

from django.core.asgi import get_asgi_application

from wahlfang.manage import setup

setup()
django_asgi_application = get_asgi_application()

from channels.auth import AuthMiddlewareStack  # pylint: disable=wrong-import-order
from channels.routing import ProtocolTypeRouter  # pylint: disable=wrong-import-order
import wahlfang.routing  # pylint: disable=wrong-import-order

application = ProtocolTypeRouter({
    "https": django_asgi_application,
    "websocket": AuthMiddlewareStack(wahlfang.routing.websocket_urlpatterns),
})



import os
import logging

from wahlfang.settings.base import *
from wahlfang.settings.wahlfang import *

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '$rl7hy0b_$*7py@t0-!^%gdlqdv0f%1+h2s%rza@=2h#1$y1vw'

DEBUG = True
# will output to your console
logging.basicConfig(
    level=logging.DEBUG,
    format='%(asctime)s %(levelname)s %(message)s',
)

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
    }
}

STATIC_ROOT = os.path.join(BASE_DIR, 'staticfiles')
MEDIA_ROOT = os.path.join(BASE_DIR, 'media')

# Mail
EMAIL_HOST = 'mail.stusta.de'
EMAIL_SENDER = 'no-reply@stusta.de'
EMAIL_PORT = 25

# LDAP
AUTH_LDAP_SERVER_URI = "ldap://ldap.stusta.de"
AUTH_LDAP_USER_DN_TEMPLATE = "cn=%(user)s,ou=account,ou=pnyx,dc=stusta,dc=mhn,dc=de"
AUTH_LDAP_START_TLS = True
AUTH_LDAP_USER_ATTR_MAP = {'email': 'mail'}
AUTH_LDAP_BIND_AS_AUTHENTICATING_USER = True



# Generated by Django 3.0.5 on 2020-05-29 19:04

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voter',
            name='first_name',
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='voter',
            name='last_name',
            field=models.CharField(blank=True, max_length=128, null=True),
        ),
        migrations.AlterField(
            model_name='voter',
            name='session',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='participants', to='vote.Session'),
        ),
        migrations.AlterField(
            model_name='voter',
            name='voter_id',
            field=models.AutoField(primary_key=True, serialize=False),
        ),
        migrations.AlterUniqueTogether(
            name='voter',
            unique_together={('session', 'email')},
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-30 19:13

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0006_auto_20200530_1836'),
    ]

    operations = [
        migrations.RenameField(
            model_name='application',
            old_name='first_name',
            new_name='display_name',
        ),
        migrations.RemoveField(
            model_name='application',
            name='last_name',
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-31 22:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0010_auto_20200601_0020'),
    ]

    operations = [
        migrations.AddField(
            model_name='voter',
            name='logged_in',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-29 12:52

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('vote', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ElectionManager',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('email', models.EmailField(max_length=254, unique=True)),
                ('sessions', models.ManyToManyField(blank=True, related_name='managers', to='vote.Session')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-21 20:04

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0016_voter_name'),
    ]

    operations = [
        migrations.AddField(
            model_name='election',
            name='send_emails_on_start',
            field=models.BooleanField(default=False),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-29 22:19

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0004_auto_20200529_2250'),
    ]

    operations = [
        migrations.AlterField(
            model_name='openvote',
            name='voter',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='open_votes', to='vote.Voter'),
        ),
    ]



# Generated by Django 3.1.2 on 2020-11-04 13:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0021_auto_20201103_1923'),
    ]

    operations = [
        migrations.AlterField(
            model_name='election',
            name='remind_text',
            field=models.TextField(blank=True, max_length=8000, null=True),
        ),
        migrations.AlterField(
            model_name='session',
            name='invite_text',
            field=models.TextField(blank=True, max_length=8000, null=True),
        ),
    ]



# Generated by Django 3.0.6 on 2020-05-30 16:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0005_auto_20200530_0019'),
    ]

    operations = [
        migrations.AlterField(
            model_name='election',
            name='max_votes_yes',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-25 20:59

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0017_election_send_emails_on_start'),
    ]

    operations = [
        migrations.AddField(
            model_name='election',
            name='remind_text',
            field=models.TextField(blank=True, max_length=1000),
        ),
        migrations.AddField(
            model_name='session',
            name='invite_text',
            field=models.TextField(blank=True, max_length=1000),
        ),
    ]



from django.core.management.base import BaseCommand

from vote.models import Voter


class Command(BaseCommand):
    help = 'Reset Voter and resend invitation'

    def add_arguments(self, parser):
        parser.add_argument('--voter_id', type=int, required=True)
        parser.add_argument('--email', type=str)
        parser.add_argument('--send-invitation', type=bool, default=True)

    def handle(self, *args, **options):
        voter_id = options['voter_id']
        voter = Voter.objects.get(pk=voter_id)
        password = voter.set_password()

        email = options['email']
        if email:
            voter.email = email
            self.stdout.write(self.style.SUCCESS('New email: "%s"' % email))

        voter.save()

        access_code = Voter.get_access_code(voter_id, password)
        if options['send_invitation']:
            voter.send_invitation(access_code, voter.session.managers.all().first().sender_email)
        self.stdout.write(self.style.SUCCESS('Successfully reset voter "%s"\nAccess Code: %s' % (voter, access_code)))



import datetime

from django.conf import settings
from django.core.management.base import BaseCommand
from django.utils import timezone

from vote.models import Election, Session


class Command(BaseCommand):
    help = 'Creates a new election'

    def add_arguments(self, parser):
        parser.add_argument('-t', '--title', type=str, required=True)
        parser.add_argument('-m', '--max-votes-yes', type=int, required=True)
        parser.add_argument('-i', '--session-id', type=int, required=True)

        # Make things a little bit easier for dev and debugging convenience
        if settings.DEBUG:
            parser.add_argument('-s', '--start-date', type=str, default=timezone.now())
            parser.add_argument('-a', '--application-due-date', type=str,
                                default=timezone.now() + datetime.timedelta(days=1))
            parser.add_argument('-e', '--end-date', type=str, default=timezone.now() + datetime.timedelta(days=2))
            parser.add_argument('-l', '--meeting-link', type=str, default="http://meeting.link")
            parser.add_argument('-d', '--meeting-time', type=str, default=timezone.now() + datetime.timedelta(days=3))
        else:
            parser.add_argument('-s', '--start-date', type=str, required=True)
            parser.add_argument('-a', '--application-due-date', type=str, required=True)
            parser.add_argument('-e', '--end-date', type=str, required=True)
            parser.add_argument('-l', '--meeting-link', type=str, required=True)
            parser.add_argument('-d', '--meeting-time', type=str, required=True)

    def handle(self, *args, **options):
        session = Session.objects.get(id=options['session_id'])
        election = Election.objects.create(
            title=options['title'],
            start_date=options['start_date'],
            end_date=options['end_date'],
            max_votes_yes=options['max_votes_yes'],
            session=session,
        )
        self.stdout.write(self.style.SUCCESS('Successfully created election "%s" with ID %i' % (election, election.id)))



import os

from django.urls import reverse_lazy

BASE_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

ADMINS = (
    ('Wahlfang Admins', 'root@localhost')
)

DEBUG = False

# export application statistics such as http request duration / latency
# will also export # of manager accounts, # of sessions, # of elections
EXPORT_PROMETHEUS_METRICS = True

ALLOWED_HOSTS = ['*']

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'crispy_forms',
    'vote',
    'management',
    'channels',
]

if EXPORT_PROMETHEUS_METRICS:
    INSTALLED_APPS += ['django_prometheus']

MIDDLEWARE = [
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'csp.middleware.CSPMiddleware',
]

if EXPORT_PROMETHEUS_METRICS:
    MIDDLEWARE = ['django_prometheus.middleware.PrometheusBeforeMiddleware'] + \
                 MIDDLEWARE + \
                 ['django_prometheus.middleware.PrometheusAfterMiddleware']

ROOT_URLCONF = 'wahlfang.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

AUTHENTICATION_BACKENDS = {
    'vote.authentication.AccessCodeBackend',
    'management.authentication.ManagementBackend',
    'django.contrib.auth.backends.ModelBackend'
}

ASGI_APPLICATION = 'wahlfang.asgi.application'

CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels.layers.InMemoryChannelLayer"
    }
}

# Password validation
# https://docs.djangoproject.com/en/3.0/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

PASSWORD_HASHERS = [
    'django.contrib.auth.hashers.Argon2PasswordHasher',
    'django.contrib.auth.hashers.PBKDF2PasswordHasher',
    'django.contrib.auth.hashers.PBKDF2SHA1PasswordHasher',
    'django.contrib.auth.hashers.BCryptSHA256PasswordHasher',
]

LOGIN_URL = reverse_lazy('vote:code_login')
LOGIN_REDIRECT_URL = reverse_lazy('vote:index')

RATELIMIT_KEY = 'ip'

# Internationalization
# https://docs.djangoproject.com/en/3.0/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'Europe/Berlin'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/3.0/howto/static-files/

STATIC_URL = '/static/'

CRISPY_TEMPLATE_PACK = 'bootstrap4'

# Content Security Policy
# https://django-csp.readthedocs.io/en/latest/configuration.html#policy-settings
CSP_DEFAULT_SRC = ("'self'",)
CSP_IMG_SRC = ("'self'", "data:",)

# File upload, etc...
MEDIA_URL = '/media/'

#: Default Logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        },
        'file': {
            'level': 'INFO',
            'class': 'logging.handlers.WatchedFileHandler',
            'filename': os.path.join(BASE_DIR, 'wahlfang.log'),
            'formatter': 'verbose',
        },
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
        },
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins', 'file'],
            'level': 'ERROR',
            'propagate': True,
        },
        'django': {
            'handlers': ['file'],
            'level': 'ERROR',
            'propagate': True,
        },
    },
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s %(process)d %(name)s %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
}



from django.core.management.base import BaseCommand, CommandError

from vote.models import Voter


class Command(BaseCommand):
    help = 'Revoke the given access code'

    def add_arguments(self, parser):
        parser.add_argument('--access_code', type=str, required=True)

    def handle(self, *args, **options):
        access_code = options['access_code']
        voter_id, password = Voter.split_access_code(access_code)
        if not voter_id:
            raise CommandError("voter not found")

        voter = Voter.objects.get(pk=voter_id)

        if not voter.check_password(password):
            raise CommandError("incorrect access_code")

        voter.set_unusable_password()
        voter.save()

        if voter.has_usable_password():
            raise CommandError("unsetting password failed")

        self.stdout.write(
            self.style.SUCCESS('Successfully revoked access for "%s"\nAccess Code: %s' % (voter, access_code)))



from django.core.management.base import BaseCommand
from django.utils import timezone

from vote.models import Election


class Command(BaseCommand):
    help = 'Process all elections and send remind emails to the voters when the elction has started'

    def add_arguments(self, parser):
        pass

    def handle(self, *args, **options):
        for election in Election.objects.all():
            if election.start_date is None or election.remind_text_sent or not election.send_emails_on_start or \
                    timezone.now() < election.start_date:
                continue
            election.remind_text_sent = True
            election.save()
            for voter in election.session.participants.all():
                voter.send_reminder(election.session.managers.all().first().sender_email, election)



from django.contrib.auth.backends import BaseBackend
from django.contrib.auth.decorators import user_passes_test

from vote.models import Voter


def voter_login_required(function=None, redirect_field_name=None):
    """
    Decorator for views that checks that the voter is logged in, redirecting
    to the log-in page if necessary.
    """
    actual_decorator = user_passes_test(
        lambda u: u.is_authenticated and isinstance(u, Voter),
        redirect_field_name=redirect_field_name
    )
    if function:
        return actual_decorator(function)
    return actual_decorator


class AccessCodeBackend(BaseBackend):
    def authenticate(self, request, **kwargs):
        access_code = kwargs.pop('access_code', None)
        if access_code is None:
            return None

        voter_id, password = Voter.split_access_code(access_code)
        if not voter_id:
            return None

        try:
            voter = Voter.objects.get(voter_id=voter_id)
        except Voter.DoesNotExist:
            # Run the default password hasher once to reduce the timing
            # difference between an existing and a nonexistent user (#20760).
            Voter().set_password(password)
        else:
            if voter.check_password(password):
                if not voter.logged_in:
                    voter.logged_in = True
                    voter.save()
                voter.backend = 'vote.authentication.AccessCodeBackend'
                return voter

        return None

    def get_user(self, user_id):
        return Voter.objects.filter(pk=user_id).first()



from datetime import timedelta, datetime

from django.test import TestCase
from django.utils import timezone
from freezegun import freeze_time

from vote.models import Election, Enc32, Voter, Session
from vote.selectors import closed_elections, open_elections, published_elections, upcoming_elections


class Enc32TestCase(TestCase):
    def test_encoding(self):
        for voter_id in (0, 12345, 999999):
            e = Enc32.encode(voter_id)
            d = Enc32.decode(e)
            self.assertEqual(voter_id, d)


class VoterTestCase(TestCase):
    def test_access_code(self):
        for voter_id in (0, 12345, 999999):
            raw_password = Enc32.alphabet
            code = Voter.get_access_code(voter_id, raw_password)
            ret_voter_id, ret_password = Voter.split_access_code(code)
            self.assertEqual(voter_id, ret_voter_id)
            self.assertEqual(raw_password, ret_password)


class ElectionSelectorsTest(TestCase):
    def test_election_selectors(self) -> None:
        now = datetime(year=2021, month=4, day=1, tzinfo=timezone.get_fixed_timezone(5))
        before = now - timedelta(seconds=5)
        bbefore = now - timedelta(seconds=10)
        after = now + timedelta(seconds=5)
        freeze_time(now).start()

        session = Session.objects.create(title="TEST")
        # upcoming elections
        all_upcoming = set()
        all_upcoming.add(Election.objects.create(session=session))
        all_upcoming.add(Election.objects.create(session=session, start_date=after))
        # open elections
        all_opened = set()
        all_opened.add(Election.objects.create(session=session, start_date=now))
        all_opened.add(Election.objects.create(session=session, start_date=before, end_date=after))
        # published elections
        all_published = set()
        all_published.add(Election.objects.create(session=session, start_date=bbefore, end_date=before,
                                                  result_published=True))
        all_published.add(Election.objects.create(session=session, start_date=before, end_date=now,
                                                  result_published=True))
        # closed (not published) elections
        all_closed = set()
        all_closed.add(Election.objects.create(session=session, start_date=bbefore, end_date=before))
        all_closed.add(Election.objects.create(session=session, start_date=before, end_date=now))

        # test upcoming
        upcoming = upcoming_elections(session)
        self.assertEqual(all_upcoming, set(upcoming))
        for e in upcoming:
            self.assertTrue(not e.started and not e.closed and not e.is_open)

        # test open
        opened = open_elections(session)
        self.assertEqual(all_opened, set(opened))
        for e in opened:
            self.assertTrue(e.started and not e.closed and e.is_open)

        # test published
        published = published_elections(session)
        self.assertEqual(all_published, set(published))
        for e in published:
            self.assertTrue(e.started and e.closed and not e.is_open and e.result_published)

        # test closed
        closed = closed_elections(session)
        self.assertEqual(all_closed, set(closed))
        for e in closed:
            self.assertTrue(e.started and e.closed and not e.is_open and not e.result_published)


def gen_data():
    session = Session.objects.create(
        title='Test session'
    )
    voter, access_code = Voter.from_data(
        email='spam@spam.spam',
        session=session
    )
    return voter, access_code



from channels.routing import URLRouter
from django.urls import path, re_path
from . import consumers

websocket_urlpatterns = URLRouter([
    path('', consumers.VoteConsumer.as_asgi()),
    re_path(r'spectator/(?P<uuid>.+)$', consumers.VoteConsumer.as_asgi()),
])



# Generated by Django 3.0.5 on 2020-05-31 22:20

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0009_remove_voter_remind_me'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='voter',
            name='first_name',
        ),
        migrations.RemoveField(
            model_name='voter',
            name='last_name',
        ),
    ]



# Generated by Django 3.0.6 on 2020-10-07 16:30

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0012_auto_20200601_0047'),
    ]

    operations = [
        migrations.AlterField(
            model_name='voter',
            name='email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
    ]



from channels.routing import URLRouter
from django.urls import re_path
from . import consumers

websocket_urlpatterns = URLRouter([
    re_path(r'election/(?P<pk>\d+)$', consumers.ElectionConsumer.as_asgi()),
    re_path(r'meeting/(?P<pk>\d+)$', consumers.SessionConsumer.as_asgi()),
])



from getpass import getpass
from secrets import token_hex

from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand
from django.core.validators import validate_email

from management.models import ElectionManager
from management.utils import is_valid_sender_email


class Command(BaseCommand):
    help = 'Create a new management login'

    def add_arguments(self, parser):
        parser.add_argument('-u', '--username', type=str, required=False)
        parser.add_argument('-e', '--email', type=str, required=False)
        parser.add_argument('--send-login-infos', action='store_true', default=False)

    def handle(self, *args, **options):
        username = options['username'] or input('Username: ')  # nosec
        email = options['email'] or input('E-Mail: ')  # nosec

        validate_email(email)
        if not is_valid_sender_email(email):
            self.stdout.write(
                self.style.ERROR(
                    f'Email must end with either of the following: {", ".join([f"@{i}" for i in settings.VALID_MANAGER_EMAIL_DOMAINS])}'))
            return

        if ElectionManager.objects.filter(email=email).exists():
            self.stdout.write(self.style.ERROR('An election manager with this email already exists'))
            return

        if options['send_login_infos']:
            password = token_hex(12)
        else:
            password = getpass('Password: ')  # nosec
            repeat_password = getpass('Repeat Password: ')  # nosec
            if password != repeat_password:
                self.stdout.write(self.style.ERROR('Passwords do not match'))
                return

        manager = ElectionManager(username=username, email=email)
        manager.set_password(password)

        if options['send_login_infos']:
            send_mail(
                f'Management login for {settings.URL}',
                f'A new management account on {settings.URL} has been created.\n'
                f'You can login under https://{settings.URL}/management '
                f'with the following data:\n\n'
                f'Username: {username}\n'
                f'Password: {password}',
                settings.EMAIL_SENDER,
                [email],
                fail_silently=False,
            )

        manager.save()
        self.stdout.write(self.style.SUCCESS(
            f'Successfully created management login with username: {username}, email: {email}'))



# Generated by Django 3.1.2 on 2020-11-11 13:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0022_auto_20201104_1438'),
    ]

    operations = [
        migrations.AddField(
            model_name='election',
            name='disable_abstention',
            field=models.BooleanField(default=False),
        ),
    ]



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin
from django.contrib.auth.forms import UserCreationForm, UserChangeForm

from management.models import ElectionManager


class ElectionManagerCreateForm(UserCreationForm):
    class Meta:
        model = ElectionManager
        fields = ('username', 'email', 'password')
        field_classes = {}


class ElectionManagerChangeForm(UserChangeForm):
    class Meta:
        model = ElectionManager
        fields = ('username', 'email',)
        field_classes = {}


class ElectionManagerAdmin(UserAdmin):
    add_form = ElectionManagerCreateForm
    form = ElectionManagerChangeForm
    model = ElectionManager
    fieldsets = (
        (None, {'fields': ('username', 'email', 'password')}),
    )
    add_fieldsets = (
        (None, {
            'classes': ('wide',),
            'fields': ('username', 'email', 'password1', 'password2')}
         ),
    )
    list_display = ('username', 'email',)
    list_filter = ('username', 'email',)
    ordering = ('username', 'email',)
    filter_horizontal = tuple()


admin.site.register(ElectionManager, ElectionManagerAdmin)



from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django import forms
from django.contrib.auth import authenticate
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from management.forms import ApplicationUploadForm
from vote.models import Application, Voter, OpenVote, VOTE_CHOICES, Vote, VOTE_ABSTENTION, VOTE_ACCEPT, \
    VOTE_CHOICES_NO_ABSTENTION


class AccessCodeAuthenticationForm(forms.Form):
    error_messages = {
        'invalid_login': _(
            "Invalid access code."
        )
    }

    access_code = forms.CharField(label='access code')

    def __init__(self, *args, request=None, **kwargs):
        """
        The 'request' parameter is set for custom auth use by subclasses.
        The form data comes in via the standard 'data' kwarg.
        """
        self.request = request
        self.user_cache = None
        super().__init__(*args, **kwargs)

        # self.fields['access_code'].max_length = 128

    def clean(self):
        access_code = self.cleaned_data.get('access_code')
        if access_code:
            self.user_cache = authenticate(self.request, access_code=access_code)
            if self.user_cache is None:
                raise forms.ValidationError(
                    self.error_messages['invalid_login'],
                    code='invalid_login',
                )

        return self.cleaned_data

    def get_user(self):
        return self.user_cache


class AvatarFileInput(forms.ClearableFileInput):
    template_name = 'vote/image_input.html'


class EmptyForm(forms.Form):
    pass


class VoteBoundField(forms.BoundField):
    def __init__(self, form, field, name, application):
        super().__init__(form, field, name)
        self.application = application


class VoteField(forms.ChoiceField):
    def __init__(self, *, application, enable_abstention=True, **kwargs):
        super().__init__(
            label=application.get_display_name(),
            choices=VOTE_CHOICES if enable_abstention else VOTE_CHOICES_NO_ABSTENTION,
            widget=forms.RadioSelect(),
            initial=VOTE_ABSTENTION if enable_abstention else None,
            **kwargs
        )
        self.application = application

    def get_bound_field(self, form, field_name):
        return VoteBoundField(form, self, field_name, application=self.application)


class VoteForm(forms.Form):
    def __init__(self, request, election, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.voter = Voter.objects.get(voter_id=request.user.voter_id)
        self.election = election
        self.request = request
        if self.election.max_votes_yes is not None:
            self.max_votes_yes = self.election.max_votes_yes
        else:
            self.max_votes_yes = self.election.applications.all().count()

        # dynamically construct form fields
        for application in self.election.applications.all():
            self.fields[f'{application.pk}'] = VoteField(application=application,
                                                         enable_abstention=self.election.enable_abstention)

        self.num_applications = self.election.applications.all().count()

    def clean(self):
        super().clean()
        if not OpenVote.objects.filter(election_id=self.election.pk, voter_id=self.voter.pk).exists():
            raise forms.ValidationError('You are not allowed to vote')

        votes_yes = 0

        for _, vote in self.cleaned_data.items():
            if vote == VOTE_ACCEPT:
                votes_yes += 1

        if votes_yes > self.max_votes_yes:
            raise forms.ValidationError(
                f'Too many "yes" votes, only max. {self.max_votes_yes} allowed.')

    def save(self, commit=True):
        votes = [
            Vote(
                election=self.election,
                candidate=Application.objects.get(pk=int(name)),
                vote=value
            ) for name, value in self.cleaned_data.items()
        ]

        # existence of can_vote object already checked in clean()
        can_vote = OpenVote.objects.get(election_id=self.election.pk, voter_id=self.voter.pk)

        if commit:
            with transaction.atomic():
                Vote.objects.bulk_create(votes)
                can_vote.delete()
            # notify manager that new votes were cast
            group = "Election-" + str(self.election.pk)
            async_to_sync(get_channel_layer().group_send)(
                group,
                {'type': 'send_reload', 'id': '#votes'}
            )

        return votes


class ApplicationUploadFormUser(ApplicationUploadForm):
    def __init__(self, election, request, *args, **kwargs):
        super().__init__(election, request, *args, **kwargs)
        if self.request.user.name:
            # these rules are meant for the StuStaNet Hausadmin election
            self.fields['display_name'].initial = self.request.user.name
            self.fields['display_name'].disabled = True
            self.fields['email'].required = True
        self.fields['email'].initial = self.request.user.email

    def save(self, commit=True):
        instance = super().save(commit=False)
        instance.voter = self.request.user
        instance.election = self.election

        if commit:
            instance.save()

        return instance



from django.conf import settings
from django.core.management.base import BaseCommand

from vote.models import Session, Voter


class Command(BaseCommand):
    help = 'Creates a new Voter'

    def add_arguments(self, parser):
        parser.add_argument('--session-id', type=int, required=True)
        parser.add_argument('--no-invitation', default=False, action='store_true')

        # Make things a little bit easier for dev and debugging convenience
        if settings.DEBUG:
            parser.add_argument('--email', type=str, default="spam@spam.spam")
        else:
            parser.add_argument('--email', type=str, required=True)

    def handle(self, *args, **options):
        session = Session.objects.get(pk=options['session_id'])

        voter, access_code = Voter.from_data(
            email=options['email'],
            session=session,
        )
        if not options['no_invitation']:
            voter.send_invitation(access_code, session.managers.all().first().sender_email)
        self.stdout.write(self.style.SUCCESS('Successfully created voter "%s"\nAccess Code: %s' % (voter, access_code)))



import json

from channels.generic.websocket import AsyncWebsocketConsumer
from channels.db import database_sync_to_async

from vote.models import Session


class VoteConsumer(AsyncWebsocketConsumer):

    async def connect(self):
        self.group = await database_sync_to_async(self.get_session_key)()  # pylint: disable=W0201
        await self.channel_layer.group_add(self.group, self.channel_name)
        await self.accept()

    async def disconnect(self, code):
        await self.channel_layer.group_discard(self.group, self.channel_name)

    async def send_reload(self, event):
        await self.send(text_data=json.dumps({
            'reload': event['id'],
        }))

    def get_session_key(self):
        if 'uuid' in self.scope['url_route']['kwargs']:
            uuid = self.scope['url_route']['kwargs']['uuid']
            session = Session.objects.get(spectator_token=uuid)
        else:
            session = self.scope['user'].session
        return "Session-" + str(session.pk)



import sys

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import authenticate, login, views as auth_views
from django.http.response import HttpResponseNotFound
from django.shortcuts import render, redirect, get_object_or_404
from django.utils.decorators import method_decorator
from ratelimit.decorators import ratelimit

from vote.authentication import voter_login_required
from vote.forms import AccessCodeAuthenticationForm, VoteForm, ApplicationUploadFormUser
from vote.models import Election, Voter, Session
from vote.selectors import open_elections, upcoming_elections, published_elections, closed_elections


class LoginView(auth_views.LoginView):
    # login view settings
    # https://docs.djangoproject.com/en/3.0/topics/auth/default/#django.contrib.auth.views.LoginView
    authentication_form = AccessCodeAuthenticationForm
    template_name = 'vote/login.html'
    redirect_authenticated_user = False

    def get(self, request, *args, **kwargs):
        print(request, args)
        u = request.user
        if u.is_authenticated and isinstance(u, Voter):
            return redirect('vote:index')
        return super().get(request, *args, **kwargs)

    @method_decorator(ratelimit(key=settings.RATELIMIT_KEY, rate='10/h', method='POST'))
    def post(self, request, *args, **kwargs):
        ratelimited = getattr(request, 'limited', False)
        if ratelimited:
            return render(request, template_name='vote/ratelimited.html', status=429)
        return super().post(request, *args, **kwargs)


@ratelimit(key=settings.RATELIMIT_KEY, rate='10/h')
def code_login(request, access_code=None):
    ratelimited = getattr(request, 'limited', False)
    if ratelimited:
        return render(request, template_name='vote/ratelimited.html', status=429)

    if not access_code:
        messages.error(request, 'No access code provided.')
        return redirect('vote:code_login')

    user = authenticate(access_code=access_code)
    if not user:
        messages.error(request, 'Invalid access code.')
        return redirect('vote:code_login')

    login(request, user)

    return redirect('vote:index')


@voter_login_required
def index(request):
    voter: Voter = request.user
    session = voter.session

    def list_elections(elections):
        return [
            (e, voter.can_vote(e), voter.has_applied(e))
            for e in elections
        ]

    context = {
        'title': session.title,
        'meeting_link': session.meeting_link,
        'voter': voter,
        'open_elections': list_elections(open_elections(session)),
        'upcoming_elections': list_elections(upcoming_elections(session)),
        'published_elections': list_elections(published_elections(session)),
        'closed_elections': list_elections(closed_elections(session)),
    }

    # overview
    return render(request, template_name='vote/index.html', context=context)


@voter_login_required
def vote(request, election_id):
    voter: Voter = request.user
    try:
        election = voter.session.elections.get(pk=election_id)
    except Election.DoesNotExist:
        return HttpResponseNotFound('Election does not exists')

    can_vote = voter.can_vote(election)
    if election.max_votes_yes is not None:
        max_votes_yes = min(election.max_votes_yes, election.applications.all().count())
    else:
        max_votes_yes = election.applications.all().count()

    context = {
        'title': election.title,
        'election': election,
        'voter': voter,
        'can_vote': can_vote,
        'max_votes_yes': max_votes_yes,
        'form': VoteForm(request, election=election)
    }

    if request.POST and can_vote:
        form = VoteForm(request, election=election, data=request.POST)
        if form.is_valid():
            form.save()
            return redirect('vote:index')

    return render(request, template_name='vote/vote.html', context=context)


@voter_login_required
def apply(request, election_id):
    voter = request.user

    election = get_object_or_404(voter.session.elections, pk=election_id)

    if not election.can_apply or not election.voters_self_apply:
        messages.add_message(request, messages.ERROR, 'Self applications are either not possible for this election or'
                                                      ' currently not accepted')
        return redirect('vote:index')

    application = voter.applications.filter(election__id=election_id)
    instance = None
    if application.exists():
        instance = application.first()

    if request.method == 'GET':
        form = ApplicationUploadFormUser(election, request, instance=instance)
    else:
        form = ApplicationUploadFormUser(election, request, data=request.POST, files=request.FILES, instance=instance)
        if form.is_valid():
            form.save()
            return redirect('vote:index')

    context = {
        'form': form,
        'election': election,
        'with_email': True,
        'with_description': True,
    }
    return render(request, template_name='vote/application.html', context=context)


@voter_login_required
def delete_own_application(request, election_id):
    voter = request.user
    election = get_object_or_404(voter.session.elections, pk=election_id)
    application = voter.applications.filter(election__id=election_id)
    if not election.can_apply:
        messages.add_message(request, messages.ERROR, 'Applications can currently not be deleted')
        return redirect('vote:index')
    if application.exists():
        instance = application.first()
        instance.delete()
        return redirect('vote:index')

    return HttpResponseNotFound('Application does not exist')


def help_page(request):
    return render(request, template_name='vote/help.html')


def spectator(request, uuid):
    session = get_object_or_404(Session.objects, spectator_token=uuid)

    context = {
        'title': session.title,
        'meeting_link': session.meeting_link,
        'open_elections': open_elections(session),
        'upcoming_elections': upcoming_elections(session),
        'published_elections': published_elections(session),
        'closed_elections': closed_elections(session),
    }
    return render(request, template_name='vote/spectator.html', context=context)




SEND_FROM_MANAGER_EMAIL = True
VALID_MANAGER_EMAIL_DOMAINS = [
    'stusta.de', 'stustanet.de', 'stusta.mhn.de', 'stusta.net', 'stusta.sexy', 'stusta.party', 'stusta.io'
]

# Base URL for template links (without 'https://')
URL = 'vote.stustanet.de'



import csv
import logging
import os
from argparse import Namespace
from functools import partial
from pathlib import Path
from typing import Dict

import qrcode
from django.conf import settings
from django.contrib import messages
from django.contrib.auth import views as auth_views
from django.http import Http404, HttpResponse
from django.http.response import HttpResponseNotFound
from django.shortcuts import render, redirect, resolve_url
from django.template.loader import get_template
from django.urls import reverse
from django.utils import timezone
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_protect
from latex.build import PdfLatexBuilder
from ratelimit.decorators import ratelimit

from management.authentication import management_login_required
from management.forms import (
    StartElectionForm,
    AddElectionForm,
    AddSessionForm,
    AddVotersForm,
    ApplicationUploadForm,
    StopElectionForm,
    AddTokensForm,
    CSVUploaderForm,
    SessionSettingsForm
)
from vote.models import Election, Application, Voter
from vote.selectors import open_elections, upcoming_elections, published_elections, closed_elections

logger = logging.getLogger('management.view')


class LoginView(auth_views.LoginView):
    # login view settings
    # https://docs.djangoproject.com/en/3.0/topics/auth/default/#django.contrib.auth.views.LoginView
    template_name = 'management/login.html'

    @method_decorator(ratelimit(key=settings.RATELIMIT_KEY, rate='10/h', method='POST'))
    def post(self, request, *args, **kwargs):
        ratelimited = getattr(request, 'limited', False)
        if ratelimited:
            return render(request, template_name='vote/ratelimited.html', status=429)
        return super().post(request, *args, **kwargs)

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url('management:index')


@management_login_required
def index(request):
    manager = request.user

    if request.GET.get("action") == "add_session":
        form = AddSessionForm(request=request, user=request.user, data=request.POST or None)
        if request.POST and form.is_valid():

            if request.POST.get("submit_type") != "test":
                ses = form.save()
                return redirect('management:session', ses.id)

            messages.add_message(request, messages.INFO, 'Test email sent.')
            Voter.send_test_invitation(
                title=form.cleaned_data['title'],
                start_date=form.cleaned_data['start_date'] if form.cleaned_data['start_date'] else timezone.now(),
                meeting_link=form.cleaned_data['meeting_link'],
                invite_text=form.cleaned_data['invite_text'],
                to_email=form.cleaned_data['email'],
                from_email=manager.sender_email
            )

        return render(request, template_name='management/add_session.html',
                      context={'form': form, 'variables': form.variables})

    context = {
        'sessions': manager.sessions.order_by('-pk')
    }
    return render(request, template_name='management/index.html', context=context)


@management_login_required
def session_detail(request, pk=None):
    manager = request.user
    session = manager.sessions.get(id=pk)
    context = {
        'session': session,
        'existing_elections': bool(session.elections),
        'open_elections': open_elections(session),
        'upcoming_elections': upcoming_elections(session),
        'published_elections': published_elections(session),
        'closed_elections': closed_elections(session),
        'voters': session.participants.all()
    }
    return render(request, template_name='management/session.html', context=context)


@management_login_required
def session_settings(request, pk=None):
    manager = request.user
    session = manager.sessions.get(pk=pk)

    form = SessionSettingsForm(instance=session, request=request, user=request.user, data=request.POST or None)
    if request.POST:
        if form.is_valid():
            if request.POST.get("submit_type") == "test":
                messages.add_message(request, messages.INFO, 'Test email sent.')
                Voter.send_test_invitation(
                    title=form.cleaned_data['title'],
                    start_date=form.cleaned_data['start_date'] if form.cleaned_data['start_date'] else timezone.now(),
                    meeting_link=form.cleaned_data['meeting_link'],
                    invite_text=form.cleaned_data['invite_text'],
                    to_email=form.cleaned_data['email'],
                    from_email=manager.sender_email
                )
            else:
                form.save()
                messages.add_message(request, messages.INFO, 'Session updated successfully!')
                return redirect('management:session', session.id)

    context = {
        'session': session,
        'elections': session.elections.order_by('pk'),
        'voters': session.participants.all(),
        'variables': form.variables,
        'form': form
    }
    return render(request, template_name='management/session_settings.html', context=context)


@management_login_required
def add_election(request, pk=None):
    manager = request.user
    session = manager.sessions.get(pk=pk)
    context = {
        'session': session,
    }

    form = AddElectionForm(session=session, request=request, user=manager, data=request.POST if request.POST else None)
    context['form'] = form
    context['variables'] = form.variables
    if request.POST and form.is_valid():
        if request.POST.get("submit_type") == "test":
            messages.add_message(request, messages.INFO, 'Test email sent.')

            test_voter = Namespace(**{
                "name": "Testname",
                "email": form.cleaned_data['email'],
            })
            test_voter.email_user = partial(Voter.email_user, test_voter)
            test_election = Namespace(**{
                "title": form.cleaned_data['title'],
                "remind_text": form.cleaned_data['remind_text'],
                "pk": 1,
                "end_date": form.cleaned_data['end_date'],
            })

            Voter.send_reminder(test_voter, manager.sender_email, test_election)
        else:
            form.save()
            return redirect('management:session', pk=session.pk)

    return render(request, template_name='management/add_election.html', context=context)


@management_login_required
def add_voters(request, pk):
    manager = request.user
    session = manager.sessions.get(pk=pk)
    context = {
        'session': session,
        'form': AddVotersForm(session=session)
    }
    form = AddVotersForm(session=session, data=request.POST if request.POST else None)
    context['form'] = form
    if request.POST and form.is_valid():
        form.save()
        return redirect('management:session', pk=pk)

    return render(request, template_name='management/add_voters.html', context=context)


@management_login_required
def add_tokens(request, pk):
    manager = request.user
    session = manager.sessions.get(pk=pk)
    context = {
        'session': session,
        'form': AddTokensForm(session=session)
    }
    form = AddTokensForm(session=session, data=request.POST if request.POST else None)
    context['form'] = form
    if request.POST and form.is_valid():
        form.save()
        return redirect('management:session', pk=pk)

    return render(request, template_name='management/add_tokens.html', context=context)


def _unpack(request, pk):
    manager = request.user
    election = Election.objects.get(pk=pk)
    session = election.session
    if not manager.sessions.filter(pk=session.pk).exists():
        raise Http404('Election does not exist/insufficient rights')
    return manager, election, session


@management_login_required
def election_detail(request, pk):
    _, election, session = _unpack(request, pk)
    context = {
        'election': election,
        'session': session,
        'applications': election.applications.all(),
        'stop_election_form': StopElectionForm(instance=election),
        'start_election_form': StartElectionForm(instance=election),
    }

    if request.POST and request.POST.get('action') == 'close' and election.is_open:
        form = StopElectionForm(instance=election, data=request.POST)
        if form.is_valid():
            form.save()
        else:
            context['stop_election_form'] = form

    if request.POST and request.POST.get('action') == 'open':
        form = StartElectionForm(instance=election, data=request.POST)
        if form.is_valid():
            form.save()
            if election.send_emails_on_start:
                for voter in session.participants.all():
                    voter.send_reminder(session.managers.all().first().sender_email, election)
        else:
            context['start_election_form'] = form

    if request.POST and request.POST.get('action') == 'publish':
        election.result_published = True
        election.save()

    return render(request, template_name='management/election.html', context=context)


@management_login_required
def election_upload_application(request, pk, application_id=None):
    _, election, _ = _unpack(request, pk)

    if not election.can_apply:
        messages.add_message(request, messages.ERROR, 'Applications are currently not accepted')
        return redirect('management:election', pk=pk)

    if application_id:
        try:
            instance = election.applications.get(pk=application_id)
        except Application.DoesNotExist:
            return HttpResponseNotFound('Application does not exist')
    else:
        instance = None

    if request.method == 'GET':
        form = ApplicationUploadForm(election, request, instance=instance)
    else:
        form = ApplicationUploadForm(election, request, data=request.POST, files=request.FILES, instance=instance)
        if form.is_valid():
            form.save()
            return redirect('management:election', election.pk)

    context = {
        'form': form,
        'election': election,
        'application_id': application_id,
        'with_email': False,
        'with_description': False,
    }
    return render(request, template_name='management/application.html', context=context)


@management_login_required
def election_delete_application(request, pk, application_id):
    e = Election.objects.filter(session__in=request.user.sessions.all(), pk=pk)
    if not e.exists():
        return HttpResponseNotFound('Election does not exist')
    e = e.first()
    try:
        a = e.applications.get(pk=application_id)
    except Application.DoesNotExist:
        return HttpResponseNotFound('Application does not exist')
    a.delete()
    return redirect('management:election', pk=pk)


@management_login_required
@csrf_protect
def delete_voter(request, pk):
    v = Voter.objects.filter(session__in=request.user.sessions.all(), pk=pk)
    if not v.exists():
        raise Http404('Voter does not exist')
    v = v.first()
    session = v.session
    v.delete()
    return redirect('management:session', pk=session.pk)


@management_login_required
@csrf_protect
def delete_election(request, pk):
    e = Election.objects.filter(session__in=request.user.sessions.all(), pk=pk)
    if not e.exists():
        return HttpResponseNotFound('Election does not exist')
    e = e.first()
    session = e.session
    e.delete()
    return redirect('management:session', pk=session.pk)


@management_login_required
@csrf_protect
def delete_session(request, pk):
    s = request.user.sessions.filter(pk=pk)
    if not s.exists():
        return HttpResponseNotFound('Session does not exist')
    s = s.first()
    s.delete()
    return redirect('management:index')


@management_login_required
def print_token(request, pk):
    session = request.user.sessions.filter(pk=pk)
    if not session.exists():
        return HttpResponseNotFound('Session does not exist')
    session = session.first()
    participants = session.participants
    tokens = [participant.new_access_token() for participant in participants.all() if participant.is_anonymous]
    if len(tokens) == 0:
        messages.add_message(request, messages.ERROR, 'No tokens have yet been generated.')
        return redirect('management:session', pk=session.pk)

    img = [qrcode.make(f'https://{settings.URL}' + reverse('vote:link_login', kwargs={'access_code': access_code}))
           for access_code in tokens]
    tmp_qr_path = '/tmp/wahlfang/qr_codes/session_{}'.format(session.pk)
    Path(tmp_qr_path).mkdir(parents=True, exist_ok=True)
    if session.meeting_link:
        meeting_qr_path = os.path.join(tmp_qr_path, 'qr_meeting.png')
        qrcode.make(session.meeting_link).save(meeting_qr_path)
    else:
        meeting_qr_path = None

    paths = []
    for idx, i in enumerate(img):
        path_i = os.path.join(tmp_qr_path, 'qr_{}.png'.format(idx))
        i.save(path_i)
        paths.append(path_i)
    zipped = [{'path': path, 'token': token} for path, token in zip(paths, tokens)]
    context = {
        'session': session,
        'tokens': zipped,
        'meeting_link_qr': meeting_qr_path
    }

    template_name = 'vote/tex/invitation.tex'
    pdf = generate_pdf(template_name, context, tmp_qr_path)
    response = HttpResponse(content_type='application/pdf')
    response['Content-Disposition'] = 'attachment; filename="tokenlist.pdf"'
    response.write(bytes(pdf))
    return response


def generate_pdf(template_name: str, context: Dict, tex_path: str):
    template = get_template(template_name).render(context).encode('utf8')
    with open("/tmp/template.tex", "wb") as f:
        f.write(template)
    pdf = PdfLatexBuilder(pdflatex='pdflatex').build_pdf(template, texinputs=[tex_path, ''])
    return pdf


@management_login_required
def import_csv(request, pk):
    session = request.user.sessions.filter(pk=pk)
    if not session.exists():
        return HttpResponseNotFound('Session does not exist')
    session = session.first()

    if request.method == 'POST':
        form = CSVUploaderForm(session, data=request.POST, files=request.FILES)
        if form.is_valid():
            form.save()
            return redirect('management:session', session.pk)
    else:
        form = CSVUploaderForm(session)
    return render(request, 'management/import_csv.html', {'form': form, 'session': session})


@management_login_required
def export_csv(request, pk):
    e = Election.objects.filter(session__in=request.user.sessions.all(), pk=pk)
    if not e.exists():
        return HttpResponseNotFound('Election does not exist')
    e = e.first()

    response = HttpResponse(content_type='text/csv')
    response['Content-Disposition'] = 'attachment; filename="results.csv"'

    writer = csv.writer(response)
    header = ['#', 'applicant', 'email', 'yes', 'no', 'abstention']
    if e.max_votes_yes is not None:
        header.append('elected')
    writer.writerow(header)
    for i, applicant in enumerate(e.election_summary):
        row = [i + 1, applicant.get_display_name(), applicant.email, applicant.votes_accept, applicant.votes_reject,
               applicant.votes_abstention]
        if e.max_votes_yes is not None:
            row.append(i < e.max_votes_yes)
        writer.writerow(row)

    return response


@management_login_required
def spectator(request, pk):
    session = request.user.sessions.filter(pk=pk)
    if not session.exists():
        return HttpResponseNotFound('Session does not exist')
    session = session.first()
    if request.POST:
        do = request.POST.get("do-type")
        if do == "create":
            session.create_spectator_token()
        elif do == "delete":
            session.spectator_token = None
            session.save()
    context = {
        'token_url': request.build_absolute_uri(
            reverse('vote:spectator', kwargs={'uuid': session.spectator_token})) if session.spectator_token else None,
        'pk': session.pk,
    }
    return render(request, template_name='management/spectator_settings.html', context=context)


def help_page(request):
    return render(request, template_name='management/help.html')



import random

from django import template

register = template.Library()


@register.simple_tag
def applicant_name(application):
    return application.get_display_name()


@register.filter
def shuffle(items):
    items = list(items)[:]
    random.shuffle(items)
    return items



import threading
import time
from typing import List, Tuple

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.conf import settings
from django.contrib.auth.base_user import AbstractBaseUser
from django.db import models

from management.utils import is_valid_sender_email
from vote.models import Session, Election, Voter


class ElectionManager(AbstractBaseUser):
    username = models.CharField(unique=True, max_length=255)
    email = models.EmailField(null=True, blank=True)
    sessions = models.ManyToManyField(Session, related_name='managers', blank=True)

    USERNAME_FIELD = 'username'
    is_staff = False
    is_superuser = False

    def __str__(self):
        return f'{self.username}'

    @property
    def sender_email(self):
        if self.email and is_valid_sender_email(self.email):
            return self.email

        return settings.EMAIL_SENDER

    def get_session(self, pk):
        return self.sessions.filter(pk=pk).first()

    def get_election(self, pk):
        return Election.objects.filter(session__in=self.sessions).filter(pk=pk).first()

    def send_invite_bulk_threaded(self, voters_codes: List[Tuple[Voter, str]]):
        def runner():
            failed_voters = list(filter(lambda i: i[0] is not None,
                                        (voter.send_invitation(code, self.sender_email) for voter, code in
                                         voters_codes)))

            def wait_heuristic():
                # heuristic sleep to avoid a channel message before the manager's websocket reconnected
                # after x send emails this sleep should be unnecessary
                if len(failed_voters) < 10:
                    time.sleep(1)

            group = "SessionAlert-" + str(voters_codes[0][0].session.pk)
            if len(failed_voters) == 0:
                wait_heuristic()
                # send message that tells the manager that all emails have been sent successfully
                async_to_sync(get_channel_layer().group_send)(
                    group,
                    {'type': 'send_succ', 'msg': "Emails send successfully!"}
                )
                return
            failed_emails_str = "".join(
                [f"<tr><td>{voter.email}</td><td>{e}</td></tr>" for voter, e in failed_voters])

            msg = 'The following email addresses failed to send and thus are probably unassigned addresses. ' \
                  'Please check them again on correctness.<table class="width100"><tr><th>Email</th>' \
                  '<th>Error</th></tr>{}</table>'

            # delete voters where the email could not be sent
            for voter, _ in failed_voters:
                voter.invalid_email = True
                voter.save()  # may trigger a lot of automatic reloads
            wait_heuristic()
            async_to_sync(get_channel_layer().group_send)(
                group,
                {'type': 'send_alert', 'msg': msg.format(failed_emails_str), 'title': 'Error during email sending',
                 'reload': '#voterCard'}
            )

        thread = threading.Thread(
            target=runner,
            args=())
        thread.start()



# Generated by Django 3.1.13 on 2021-08-06 08:40

from django.db import migrations
from django.db.models import Q


def update_values(apps, schema_editor):
    election = apps.get_model('vote', 'election')
    election.objects.update(result_published=Q(result_published=False))
    election.objects.update(enable_abstention=Q(enable_abstention=False))


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0028_auto_20210804_2335'),
    ]

    operations = [
        migrations.RenameField(
            model_name='election',
            old_name='result_unpublished',
            new_name='result_published',
        ),
        migrations.RenameField(
            model_name='election',
            old_name='disable_abstention',
            new_name='enable_abstention',
        ),
        migrations.RunPython(update_values, reverse_code=update_values),
    ]



from django.db.models import Q
from django.utils import timezone

from vote.models import Election, Session


def upcoming_elections(session: Session):
    return Election.objects.filter(session=session).filter(
        Q(start_date__gt=timezone.now()) | Q(start_date__isnull=True)
    ).order_by('start_date')


def open_elections(session: Session):
    return Election.objects.filter(session=session).filter(
        Q(start_date__isnull=False, end_date__isnull=False, start_date__lte=timezone.now(), end_date__gt=timezone.now())
        | Q(start_date__isnull=False, end_date__isnull=True, start_date__lte=timezone.now())
    ).order_by('-start_date')


def _closed_elections(session: Session):
    return Election.objects.filter(session=session).filter(
        Q(end_date__lte=timezone.now(), end_date__isnull=False)
    ).order_by('-start_date')


def published_elections(session: Session):
    return _closed_elections(session).filter(result_published=True)


def closed_elections(session: Session):
    return _closed_elections(session).filter(result_published=False)



# Generated by Django 3.1.2 on 2020-10-12 19:07

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0015_auto_20201011_1826'),
    ]

    operations = [
        migrations.AddField(
            model_name='voter',
            name='name',
            field=models.CharField(blank=True, max_length=256, null=True),
        ),
    ]



from channels.routing import URLRouter
from django.urls import path

import management.routing
import vote.routing

websocket_urlpatterns = URLRouter([
    path('', vote.routing.websocket_urlpatterns),
    path('management/', management.routing.websocket_urlpatterns),
])



import os
import sys
import textwrap
import uuid
from argparse import Namespace
from datetime import datetime
from functools import partial
from io import BytesIO
from typing import Tuple, Optional

import PIL
from PIL import Image
from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.conf import settings
from django.contrib.auth import password_validation
from django.contrib.auth.hashers import (
    check_password, is_password_usable, make_password,
)
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.mail import send_mail
from django.db import models
from django.db.models import Count, Q, CASCADE
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.html import strip_tags

VOTE_ACCEPT = 'accept'
VOTE_ABSTENTION = 'abstention'
VOTE_REJECT = 'reject'
VOTE_CHOICES = [
    (VOTE_ABSTENTION, 'Abstention'),
    (VOTE_ACCEPT, 'Yes'),
    (VOTE_REJECT, 'No'),
]

VOTE_CHOICES_NO_ABSTENTION = [
    (VOTE_ACCEPT, 'Yes'),
    (VOTE_REJECT, 'No'),
]


class Enc32:
    alphabet = "0123456789abcdefghjknpqrstuvwxyz"
    dec_map = {}
    for index, c in enumerate(alphabet):
        dec_map[c] = index
    dec_map['o'] = dec_map['0']
    dec_map['l'] = dec_map['1']
    dec_map['i'] = dec_map['1']
    dec_map['m'] = dec_map['n']

    @staticmethod
    def encode(i, length=None):
        i = int(i)
        out = ""
        while i > 0:
            idx = i & 0x1f
            out = Enc32.alphabet[idx] + out
            i >>= 5
        if length:
            if len(out) > length:
                raise ValueError("value too large for given length")
            out = Enc32.alphabet[0] * (length - len(out)) + out
        return out

    @staticmethod
    def decode(s):
        i = 0
        for c in s:
            i <<= 5
            i += Enc32.dec_map[c]
        return i


class Session(models.Model):
    title = models.CharField(max_length=256)
    meeting_link = models.CharField(max_length=512, blank=True, null=True)
    start_date = models.DateTimeField(blank=True, null=True)
    invite_text = models.TextField(max_length=8000, blank=True, null=True)
    spectator_token = models.UUIDField(unique=True, null=True, blank=True)

    def create_spectator_token(self):
        myid = uuid.uuid4()
        while Session.objects.filter(spectator_token=myid).exists():
            myid = uuid.uuid4()
        self.spectator_token = myid
        self.save()
        return myid


class Election(models.Model):
    title = models.CharField(max_length=512)
    start_date = models.DateTimeField(blank=True, null=True)
    end_date = models.DateTimeField(blank=True, null=True)
    max_votes_yes = models.IntegerField(blank=True, null=True)
    session = models.ForeignKey(Session, related_name='elections', on_delete=CASCADE)
    result_published = models.BooleanField(null=False, default=False)
    enable_abstention = models.BooleanField(default=True)
    voters_self_apply = models.BooleanField(default=False)
    send_emails_on_start = models.BooleanField(default=False)
    remind_text = models.TextField(max_length=8000, blank=True, null=True)
    remind_text_sent = models.BooleanField(default=False)

    @property
    def started(self):
        if self.start_date is not None:
            return timezone.now() >= self.start_date

        return False

    @property
    def closed(self):
        if self.end_date:
            return self.end_date <= timezone.now()

        return False

    @property
    def is_open(self):
        if self.start_date and self.end_date:
            return self.start_date <= timezone.now() < self.end_date

        if self.start_date:
            return self.start_date <= timezone.now()

        return False

    @property
    def can_apply(self):
        if self.start_date:
            return timezone.now() < self.start_date

        return True

    @property
    def election_summary(self):
        if not self.closed:
            return self.objects.none()

        votes_accept = Count('votes', filter=Q(votes__vote=VOTE_ACCEPT))
        votes_reject = Count('votes', filter=Q(votes__vote=VOTE_REJECT))
        votes_abstention = Count('votes', filter=Q(votes__vote=VOTE_ABSTENTION))

        applications = Application.objects.filter(election_id=self.pk).annotate(
            votes_accept=votes_accept,
            votes_reject=votes_reject,
            votes_abstention=votes_abstention
        ).order_by('-votes_accept')

        return applications

    def number_voters(self):
        return self.session.participants.count()

    def number_votes_open(self):
        return self.open_votes.count()

    def number_votes_cast(self):
        if self.applications.all().count() == 0:
            return 0
        return int(self.votes.count() / self.applications.all().count())

    def save(self, force_insert=False, force_update=False, using=None, update_fields=None):
        super().save(force_insert, force_update, using, update_fields)
        # notify users to reload their page
        group = "Session-" + str(self.session.pk)
        async_to_sync(get_channel_layer().group_send)(
            group,
            {'type': 'send_reload', 'id': '#electionCard'}
        )

    def __str__(self):
        return self.title


class Voter(models.Model):
    voter_id = models.AutoField(primary_key=True)
    password = models.CharField(max_length=256)
    email = models.EmailField(null=True, blank=True)
    invalid_email = models.BooleanField(default=False)
    session = models.ForeignKey(Session, related_name='participants', on_delete=models.CASCADE)
    logged_in = models.BooleanField(default=False)
    name = models.CharField(max_length=256, blank=True, null=True)

    # Stores the raw password if set_password() is called so that it can
    # be passed to password_changed() after the model is saved.
    _password = None

    USERNAME_FIELD = 'voter_id'

    class Meta:
        unique_together = ('session', 'email')

    def __str__(self):
        if self.email is None:
            return f'anonymous-{self.pk}'

        return self.email

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if update_fields == ['last_login']:
            return

        super().save(force_insert, force_update, using, update_fields)
        if self._password is not None:
            password_validation.password_changed(self._password, self)
            self._password = None
        # notify manager to reload their page, if the user logged in
        group = "Login-Session-" + str(self.session.pk)
        async_to_sync(get_channel_layer().group_send)(
            group,
            {'type': 'send_reload', 'id': '#voterCard'}
        )

    def set_password(self, raw_password=None):
        if not raw_password:
            raw_password = get_random_string(length=20, allowed_chars=Enc32.alphabet)
        self.password = make_password(raw_password)
        self._password = raw_password
        return raw_password

    def check_password(self, raw_password):
        """
        Return a boolean of whether the raw_password was correct. Handles
        hashing formats behind the scenes.
        """

        def setter(raw_password):
            self.set_password(raw_password)
            # Password hash upgrades shouldn't be considered password changes.
            self._password = None
            self.save(update_fields=['password'])

        return check_password(raw_password, self.password, setter)

    def set_unusable_password(self):
        # Set a value that will never be a valid hash
        self.password = make_password(None)

    def has_usable_password(self):
        """
        Return False if set_unusable_password() has been called for this user.
        """
        return is_password_usable(self.password)

    def clean(self):
        setattr(self, self.email, self.normalize_email(self.email))

    @classmethod
    def normalize_email(cls, email):
        """
        Normalize the email address by lowercasing the domain part of it.
        """
        email = email or ''
        try:
            email_name, domain_part = email.strip().rsplit('@', 1)
        except ValueError:
            pass
        else:
            email = email_name + '@' + domain_part.lower()
        return email

    def email_user(self, subject, message, from_email=None, **kwargs) -> Tuple[Optional['Voter'], Optional[str]]:
        """Send an email to this user."""
        if self.email is not None:
            try:
                send_mail(subject, message, from_email, [self.email], **kwargs)
            except Exception as e:  # pylint: disable=W0703
                return self, str(e)
        # None means everything is ok
        return None, None

    @property
    def is_authenticated(self):
        """
        Always return True. This is a way to tell if the user has been
        authenticated in templates.
        """
        return True

    @property
    def is_active(self):
        return self.has_usable_password()

    def can_vote(self, election: Election):
        return election.is_open and OpenVote.objects.filter(voter_id=self.voter_id, election_id=election.id).exists()

    def has_applied(self, election: Election):
        return self.applications.filter(election=election).exists()

    @property
    def is_staff(self):
        return False

    @property
    def is_anonymous(self):
        return self.email is None

    def has_module_perms(self, app_label):
        return False

    def get_username(self):
        return str(self)

    @staticmethod
    def send_test_invitation(title: str, invite_text: str, start_date: datetime, meeting_link: str, to_email: str,
                             from_email: str):
        test_session = Namespace(**{
            "title": title,
            "invite_text": invite_text,
            "start_date": start_date,
            'meeting_link': meeting_link,
        })

        test_voter = Namespace(**{
            "name": "Testname",
            "email": to_email,
            "session": test_session,
        })
        test_voter.email_user = partial(Voter.email_user, test_voter)

        Voter.send_invitation(test_voter, "mock-up-access-token", from_email)

    def send_invitation(self, access_code: str, from_email: str) -> Tuple[Optional['Voter'], Optional[str]]:
        if not self.email:
            return None, None
        subject = f'Invitation for {self.session.title}'
        if self.session.invite_text:
            if self.session.start_date:
                # cast to correct time zone
                current_tz = timezone.get_current_timezone()
                st = current_tz.normalize(self.session.start_date)
            context = {
                'name': self.name,
                'title': self.session.title,
                'access_code': access_code,
                'login_url': f'https://{settings.URL}' + reverse('vote:link_login',
                                                                 kwargs={'access_code': access_code}),
                'start_date': st.strftime("%d.%m.%Y") if self.session.start_date else "",
                'start_time': st.strftime("%H:%M") if self.session.start_date else "",
                'start_date_en': st.strftime("%Y/%m/%d") if self.session.start_date else "",
                'start_time_en': st.strftime("%I:%M %p") if self.session.start_date else "",
                'base_url': f'https://{settings.URL}',
                'meeting_link': self.session.meeting_link
            }
            body_html = self.session.invite_text.format(**context)
        else:
            context = {
                'voter': self,
                'session': self.session,
                'base_url': f'https://{settings.URL}',
                'login_url': f'https://{settings.URL}' + reverse('vote:link_login',
                                                                 kwargs={'access_code': access_code}),
                'access_code': access_code,
            }
            body_html = render_to_string('vote/mails/invitation.j2', context=context)

        return self.email_user(
            subject=subject,
            message=strip_tags(body_html),
            from_email=from_email,
            html_message=body_html.replace('\n', '<br/>'),
            fail_silently=False
        )

    def send_reminder(self, from_email: str, election):
        if not self.email:
            return
        subject = f'{election.title} is now open'
        if election.remind_text:
            if election.end_date:
                # cast to correct time zone
                current_tz = timezone.get_current_timezone()
                et = current_tz.normalize(election.end_date)
            context = {
                'name': self.name,
                'title': election.title,
                'url': f'https://{settings.URL}' + reverse('vote:vote', kwargs={'election_id': election.pk}),
                'end_date': et.strftime("%d.%m.%y") if election.end_date else "",
                'end_time': et.strftime("%H:%M") if election.end_date else "",
                'end_date_en': et.strftime("%Y/%m/%d") if election.end_date else "",
                'end_time_en': et.strftime("%I:%M %p") if election.end_date else "",
            }
            body_html = election.remind_text.format(**context)
        else:
            context = {
                'voter': self,
                'election': election,
                'url': f'https://{settings.URL}' + reverse('vote:vote', kwargs={'election_id': election.pk}),
            }
            body_html = render_to_string('vote/mails/start.j2', context=context)

        self.email_user(
            subject=subject,
            message=strip_tags(body_html),
            from_email=from_email,
            html_message=body_html.replace('\n', '<br/>'),
            fail_silently=True
        )

    @staticmethod
    def get_access_code(voter, raw_password):
        if isinstance(voter, Voter):
            voter_id = voter.voter_id
        else:
            voter_id = int(voter)

        enc_id = Enc32.encode(voter_id, 4)
        return '-'.join(textwrap.wrap(enc_id + raw_password, 6))

    @staticmethod
    def split_access_code(access_code=None):
        if not access_code:
            return None, None

        access_code = access_code.replace('-', '')
        if len(access_code) < 5 or not all(c in Enc32.alphabet for c in access_code):
            return None, None

        voter_id = Enc32.decode(access_code[:4])
        password = access_code[4:].lower()
        return voter_id, password

    @classmethod
    def from_data(cls, session, email=None, name=None) -> Tuple['Voter', str]:
        voter = Voter(
            session=session,
            email=email,
            name=name,
        )
        password = voter.set_password()
        voter.save()

        # add open elections from the session where the user was added
        open_votes = [OpenVote(election=election, voter=voter) for election in session.elections.all()
                      if not election.closed]
        OpenVote.objects.bulk_create(open_votes)

        return voter, cls.get_access_code(voter.voter_id, password)

    def new_access_token(self):
        password = self.set_password()
        self.logged_in = False
        self.save()
        return self.get_access_code(self, password)


def avatar_file_name(instance, filename):
    ext = filename.split('.')[-1]
    return os.path.join('avatars', str(uuid.uuid4()) + '.' + ext)


class Application(models.Model):
    text = models.TextField(max_length=250, blank=True)
    avatar = models.ImageField(upload_to=avatar_file_name, null=True, blank=True)
    election = models.ForeignKey(Election, related_name='applications', on_delete=models.CASCADE)
    display_name = models.CharField(max_length=256)
    email = models.EmailField(null=True, blank=True)
    voter = models.ForeignKey(Voter, related_name="applications", null=True, blank=True, on_delete=models.CASCADE)

    _old_avatar = None

    class Meta:
        unique_together = ('voter', 'election')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._old_avatar = self.avatar

    def __str__(self):
        return f'Application of {self.get_display_name()} for {self.election}'

    def get_display_name(self):
        return f'{self.display_name}'

    def save(self, force_insert=False, force_update=False, using=None,
             update_fields=None):
        if self.avatar and self._old_avatar != self.avatar:
            # remove old file
            if self._old_avatar and os.path.isfile(self._old_avatar.path):
                # let's not play russian roulette
                path = os.path.normpath(self._old_avatar.path)
                if path.startswith(os.path.join(settings.MEDIA_ROOT, 'avatars')):
                    os.remove(path)

            max_width = 100
            max_height = 100
            img = Image.open(self.avatar)

            # remove alpha channel
            if img.mode in ('RGBA', 'LA'):
                background = Image.new(img.mode[:-1], img.size, '#FFF')
                background.paste(img, img.split()[-1])
                img = background

            # resize
            width = max_width
            width_percent = (width / float(img.size[0]))
            height = int((float(img.size[1]) * float(width_percent)))
            if height > max_height:
                height = max_height
                height_percent = (height / float(img.size[1]))
                width = int((float(img.size[0]) * float(height_percent)))
            img = img.resize((width, height), PIL.Image.ANTIALIAS)

            output = BytesIO()
            img.save(output, format='JPEG', quality=95)
            output.seek(0)
            self.avatar = InMemoryUploadedFile(output, 'ImageField', '%s.jpg' % self.avatar.name.split('.')[0],
                                               'image/jpeg', sys.getsizeof(output), None)
            self._old_avatar = self.avatar

        super().save(force_insert, force_update, using, update_fields)


class OpenVote(models.Model):
    election = models.ForeignKey(Election, related_name='open_votes', on_delete=models.CASCADE)
    voter = models.ForeignKey(Voter, related_name='open_votes', on_delete=models.CASCADE)

    def can_vote(self, voter_id, election_id):
        return self.objects.filter(voter_id=voter_id, election_id=election_id).exists()


class Vote(models.Model):
    election = models.ForeignKey(Election, related_name='votes', on_delete=models.CASCADE)
    candidate = models.ForeignKey(Application, related_name='votes', on_delete=models.CASCADE)
    vote = models.CharField(choices=VOTE_CHOICES, max_length=max(len(x[0]) for x in VOTE_CHOICES))
    # save method is not called on bulk_create in forms.VoteForm.
    # The model update listener for websockets is implemented in the form.



from django.contrib.auth import views as auth_views
from django.urls import path
from prometheus_client import Gauge

import vote.views
from management import views
from management.models import ElectionManager
from vote.models import Election, Session

app_name = 'management'

election_gauge = Gauge('wahlfang_election_count', 'Wahlfang Number of Elections')
election_gauge.set_function(lambda: Election.objects.all().count())

election_manager_gauge = Gauge('wahlfang_election_manager_count', 'Wahlfang Number of Election Managers')
election_manager_gauge.set_function(lambda: ElectionManager.objects.all().count())

session_gauge = Gauge('wahlfang_session_count', 'Wahlfang Number of Sessions')
session_gauge.set_function(lambda: Session.objects.all().count())

urlpatterns = [
    path('', views.index, name='index'),
    path('help', views.help_page, name='help'),

    # Session
    path('meeting/<int:pk>', views.session_detail, name='session'),
    path('meeting/<int:pk>/settings', views.session_settings, name='session_settings'),
    path('meeting/<int:pk>/delete_session', views.delete_session, name='delete_session'),
    path('meeting/<int:pk>/add_voters', views.add_voters, name='add_voters'),
    path('meeting/<int:pk>/add_tokens', views.add_tokens, name='add_tokens'),
    path('meeting/<int:pk>/add_election', views.add_election, name='add_election'),
    path('meeting/<int:pk>/print_token', views.print_token, name='print_token'),
    path('meeting/<int:pk>/import_csv', views.import_csv, name='import_csv'),
    path('meeting/<int:pk>/spectator', views.spectator, name='spectator'),

    # Election
    path('election/<int:pk>/add_application', views.election_upload_application, name='add_application'),
    path('election/<int:pk>/edit/<int:application_id>', views.election_upload_application, name='edit_application'),
    path('election/<int:pk>/edit/<int:application_id>/delete_application', views.election_delete_application,
         name='delete_application'),
    path('election/<int:pk>', views.election_detail, name='election'),
    path('election/<int:pk>/delete_voter', views.delete_voter, name='delete_voter'),
    path('election/<int:pk>/delete_election', views.delete_election, name='delete_election'),
    path('election/<int:pk>/export_csv', views.export_csv, name='export_csv'),

    # account management stuff
    path('login', views.LoginView.as_view(), name='login'),
    path('logout', auth_views.LogoutView.as_view(
        next_page='management:login',
    ), name='logout')
]



# Generated by Django 3.0.5 on 2020-05-31 21:31

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0008_election_result_published'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='voter',
            name='remind_me',
        ),
    ]



# Generated by Django 3.0.5 on 2020-05-29 20:50

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0003_auto_20200529_2118'),
    ]

    operations = [
        migrations.AlterField(
            model_name='election',
            name='session',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='elections', to='vote.Session'),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-07 14:36

from django.db import migrations, models


def copy_house_id(apps, schema_editor):
    ElectionManager = apps.get_model('management', 'ElectionManager')
    for e in ElectionManager.objects.all():
        e.username = e.email
        e.save()


class Migration(migrations.Migration):

    dependencies = [
        ('management', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='electionmanager',
            name='username',
            field=models.CharField(null=True, max_length=255),
        ),
        migrations.RunPython(code=copy_house_id, reverse_code=migrations.RunPython.noop),
        migrations.AlterField(
            model_name='electionmanager',
            name='username',
            field=models.CharField(max_length=255, unique=True),
        ),
        migrations.AlterField(
            model_name='electionmanager',
            name='email',
            field=models.EmailField(blank=True, max_length=254, null=True),
        ),
    ]



# Generated by Django 3.1.2 on 2020-11-03 18:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('vote', '0020_election_remind_text_sent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='session',
            name='start_date',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



import json

from channels.generic.websocket import AsyncWebsocketConsumer


class ElectionConsumer(AsyncWebsocketConsumer):

    async def connect(self):
        self.group = "Election-" + self.scope['url_route']['kwargs']['pk']  # pylint: disable=W0201
        await self.channel_layer.group_add(self.group, self.channel_name)
        await self.accept()

    async def disconnect(self, code):
        await self.channel_layer.group_discard(self.group, self.channel_name)

    async def send_reload(self, event):
        await self.send(text_data=json.dumps({
            'reload': event['id'],
        }))


class SessionConsumer(AsyncWebsocketConsumer):

    async def connect(self):
        session = self.scope['url_route']['kwargs']['pk']
        # reload if a new voter logged in, or if a election of the session was changed (like added)
        self.groups = ["Login-Session-" + session, "Session-" + session,
                       "SessionAlert-" + session]  # pylint: disable=W0201
        for group in self.groups:
            await self.channel_layer.group_add(group, self.channel_name)
        await self.accept()

    async def disconnect(self, code):
        for group in self.groups:
            await self.channel_layer.group_discard(group, self.channel_name)

    async def send_reload(self, event):
        await self.send(text_data=json.dumps({
            'reload': event['id'],
        }))

    async def send_alert(self, event):
        await self.send(text_data=json.dumps({
            'alert': {'title': event.get('title', 'Alert'), 'msg': event['msg'], 'reload': event.get('reload')},
        }))

    async def send_succ(self, event):
        await self.send(text_data=json.dumps({
            'succ': event['msg'],
        }))



# Generated by Django 3.1.2 on 2020-10-11 16:26

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('vote', '0014_election_voters_self_apply'),
    ]

    operations = [
        migrations.AddField(
            model_name='application',
            name='voter',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='application', to='vote.voter'),
        ),
        migrations.AlterUniqueTogether(
            name='application',
            unique_together={('voter', 'election')},
        ),
    ]
